@@ -156,23 +156,38 @@ void TableState::pushLastCommit(const std::string &code,
     TABLE_DEBUG() << "TableState::pushLastCommit " << lastSegment
                   << " code: " << code;
     constexpr size_t limit = 10;
-    const auto length = utf8::length(lastSegment);
-    if (length == 1) {
-        lastSingleCharCommit_.emplace_back(code, lastSegment);
-        while (lastSingleCharCommit_.size() > limit) {
-            lastSingleCharCommit_.pop_front();
+    const auto length = utf8::lengthValidated(lastSegment);
+    // Sanity check.
+    if (length <= 0 || length == utf8::INVALID_LENGTH) {
+        return;
+    }
+
+    if (length == 1 || *context_->config().autoPhraseWithPhrase) {
+        // Single character is with code as hint
+        if (length == 1) {
+            autoPhraseBuffer_.emplace_back(code, lastSegment);
+        } else {
+            auto range = fcitx::utf8::MakeUTF8CharRange(lastSegment);
+            for (auto iter = std::begin(range); iter != std::end(range);
+                 iter++) {
+                autoPhraseBuffer_.emplace_back("", iter.view());
+            }
+        }
+        while (autoPhraseBuffer_.size() > limit) {
+            autoPhraseBuffer_.pop_front();
         }
 
         std::string singleCharString;
         std::vector<std::string> codeHints;
-        for (const auto &[code, chr] : lastSingleCharCommit_) {
+        for (const auto &[code, chr] : autoPhraseBuffer_) {
             singleCharString += chr;
             codeHints.push_back(code);
         }
-        TABLE_DEBUG() << "learnAutoPhrase " << singleCharString << codeHints;
+        TABLE_DEBUG() << "learnAutoPhrase " << autoPhraseBuffer_ << " "
+                      << singleCharString << codeHints;
         context_->learnAutoPhrase(singleCharString, codeHints);
     } else {
-        lastSingleCharCommit_.clear();
+        autoPhraseBuffer_.clear();
     }
 
     if (length == 1) {
@@ -301,7 +316,8 @@ bool TableState::handleCandidateList(const TableConfig &config,
 
 bool TableState::handlePinyinMode(KeyEvent &event) {
     auto *context = context_.get();
-    const auto &pinyinKey = *context->config().pinyinKey;
+    const auto &config = context_->config();
+    const auto &pinyinKey = *config.pinyinKey;
     if (pinyinKey.sym() == FcitxKey_None) {
         return false;
     }
@@ -364,7 +380,6 @@ bool TableState::handlePinyinMode(KeyEvent &event) {
     if (!pinyinModeBuffer_.empty()) {
         const auto &dict = engine_->pinyinDict();
         const auto &lm = engine_->pinyinModel();
-        const auto &config = context_->config();
         auto pinyin = libime::PinyinEncoder::encodeOneUserPinyin(
             pinyinModeBuffer_.userInput());
 
@@ -409,7 +424,11 @@ bool TableState::handlePinyinMode(KeyEvent &event) {
     preeditText.append(pinyinModePrefix_);
     preeditText.append(pinyinModeBuffer_.userInput());
     if (ic_->capabilityFlags().test(CapabilityFlag::Preedit)) {
-        preeditText.setCursor(0);
+        if (*config.preeditCursorPositionAtBeginning) {
+            preeditText.setCursor(0);
+        } else {
+            preeditText.setCursor(preeditText.textLength());
+        }
         inputPanel.setClientPreedit(preeditText);
     } else {
         preeditText.setCursor(preeditText.textLength());
@@ -735,7 +754,19 @@ void TableState::keyEvent(const InputMethodEntry &entry, KeyEvent &event) {
         return;
     }
     if (!event.key().hasModifier() && chr && context->isValidInput(chr)) {
-
+        auto candidateList = ic_->inputPanel().candidateList();
+        auto autoSelectHint = 0;
+        if (candidateList && candidateList->size()) {
+            int idx = candidateList->cursorIndex();
+            if (idx >= 0) {
+                auto cand = dynamic_cast<const TableCandidateWord *>(
+                    &candidateList->candidate(idx));
+                if (cand) {
+                    autoSelectHint = cand->idx_;
+                }
+            }
+        }
+        context->setAutoSelectIndex(autoSelectHint);
         {
             CommitAfterSelectWrapper commitAfterSelectRAII(this);
             context->type(str);
